/*
 * File         : mpi_Request.c
 * Headerfile   : mpi_Request.h 
 * Author       : Sung-Hoon Ko, Xinying Li, Bryan Carpenter
 * Created      : Thu Apr  9 12:22:15 1998
 * Revision     : $Revision: 1.5 $
 * Updated      : $Date: 1999/10/06 21:39:47 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

#include <mpi.h>
#include "mpi_Request.h"

jfieldID reqhandleID;

extern jfieldID stathandleID ;
extern jfieldID sourceID, tagID, countID, indexID ;
extern jmethodID handleConstructorID ;

jfieldID bufSaveID, bufptrSaveID, baseTypeSaveID, offsetSaveID ;

extern void releaseBufPtr(JNIEnv *env,jobject buf, int baseType,
                          void* bufptr, int offset) ;


/* extern jclass status_class ; */


/*
 * Class:     mpi_Status
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Request_init 
  (JNIEnv *env, jclass thisClass)
{
  reqhandleID    = (*env)->GetFieldID(env, thisClass, "handle", "J") ;
  bufSaveID      = (*env)->GetFieldID(env, thisClass, "bufSave",
                                      "Ljava/lang/Object;") ;
  bufptrSaveID   = (*env)->GetFieldID(env, thisClass, "bufptrSave", "J") ;
  baseTypeSaveID = (*env)->GetFieldID(env, thisClass, "baseTypeSave", "I") ;
  offsetSaveID   = (*env)->GetFieldID(env, thisClass, "offsetSave", "I") ;
}                      

/*
 * Class:     mpi_Request
 * Method:    GetReq
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_mpi_Request_GetReq
  (JNIEnv *env, jobject jthis, jint type)
{
  switch (type) {
    case 0:
      (*env)->SetLongField(env,jthis,reqhandleID,(jlong)MPI_REQUEST_NULL);
      break;
	default:
	  break;
  }
}

/*
 * Class:     mpi_Request
 * Method:    Cancel
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Request_Cancel
  (JNIEnv *env, jobject jthis)
{
   MPI_Request req;
   req=(MPI_Request)((*env)->GetLongField(env,jthis,reqhandleID));
   MPI_Cancel(&req);
}

/*
 * Class:     mpi_Request
 * Method:    Free
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Request_Free
  (JNIEnv *env, jobject jthis)
{
   MPI_Request req;
   req=(MPI_Request)((*env)->GetLongField(env,jthis,reqhandleID));
   MPI_Request_free(&req);
}

/*
 * Class:     mpi_Request
 * Method:    Is_null
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_mpi_Request_Is_1null
  (JNIEnv *env, jobject jthis)
{
  MPI_Request req;
  req=(MPI_Request)((*env)->GetLongField(env,jthis,reqhandleID));
  if(req==MPI_REQUEST_NULL)
    return JNI_TRUE;
  else 
    return JNI_FALSE;
}

/*
 * Class:     mpi_Request
 * Method:    Wait
 * Signature: (Lmpi/Status;)Lmpi/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_Wait
  (JNIEnv *env, jobject jthis, jobject stat)
{
  MPI_Request req =
      (MPI_Request)((*env)->GetLongField(env,jthis,reqhandleID)) ;

  jobject buf  = (*env)->GetObjectField(env, jthis, bufSaveID) ;
  void* bufptr = (void*) (*env)->GetLongField(env, jthis, bufptrSaveID) ;
  int baseType = (*env)->GetIntField(env, jthis, baseTypeSaveID) ;
  int offset   = (*env)->GetIntField(env, jthis, offsetSaveID) ;

  MPI_Status *status =
      (MPI_Status *)((*env)->GetLongField(env,stat,stathandleID));

  MPI_Wait(&req, status);

  (*env)->SetLongField(env,jthis,reqhandleID,(jlong)req);

  releaseBufPtr(env, buf, baseType, bufptr, offset);

  (*env)->SetIntField(env,stat, sourceID,status->MPI_SOURCE);
  (*env)->SetIntField(env,stat, tagID,status->MPI_TAG);

  return stat;
}

/*
 * Class:     mpi_Request
 * Method:    Test
 * Signature: (Lmpi/Status;)Lmpi/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_Test
  (JNIEnv *env, jobject jthis, jobject stat)
{
  int flag;
  MPI_Request req = (MPI_Request)((*env)->GetLongField(env,jthis,reqhandleID));
  MPI_Status *status =
      (MPI_Status *)((*env)->GetLongField(env,stat,stathandleID));

  MPI_Test(&req, &flag, status);

  (*env)->SetLongField(env,jthis,reqhandleID,(jlong)req);

  if(flag) {
    jobject buf  = (*env)->GetObjectField(env, jthis, bufSaveID) ;
    void* bufptr = (void*) (*env)->GetLongField(env, jthis, bufptrSaveID) ;
    int baseType = (*env)->GetIntField(env, jthis, baseTypeSaveID) ;
    int offset   = (*env)->GetIntField(env, jthis, offsetSaveID) ;

    releaseBufPtr(env, buf, baseType, bufptr, offset);

    (*env)->SetIntField(env,stat, sourceID,status->MPI_SOURCE);
    (*env)->SetIntField(env,stat, tagID,status->MPI_TAG);

    return stat;
  }
  else
    return NULL;
}

/*
 * Class:     mpi_Request
 * Method:    Waitany
 * Signature: ([Lmpi/Request;Lmpi/Status;)Lmpi/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_Waitany
  (JNIEnv *env, jclass jthis, jobjectArray array_of_request, jobject stat)
{
  int i,index;
  int count=(*env)->GetArrayLength(env,array_of_request);
  MPI_Request *reqs=(MPI_Request*)calloc(count, sizeof(MPI_Request));

  jobject buf  = (*env)->GetObjectField(env, jthis, bufSaveID) ;
  void* bufptr = (void*) (*env)->GetLongField(env, jthis, bufptrSaveID) ;
  int baseType = (*env)->GetIntField(env, jthis, baseTypeSaveID) ;
  int offset   = (*env)->GetIntField(env, jthis, offsetSaveID) ;

  MPI_Status *status =
      (MPI_Status *)((*env)->GetLongField(env,stat,stathandleID));

  for(i=0; i<count; i++)
    reqs[i]=(MPI_Request)((*env)->GetLongField(env,
	       (*env)->GetObjectArrayElement(env,array_of_request,i),
               reqhandleID)) ;

  MPI_Waitany(count, reqs, &index, status);

  for(i=0; i<count; i++)
    (*env)->SetLongField(env,
              (*env)->GetObjectArrayElement(env,array_of_request,i),
              reqhandleID, (jlong) reqs[i]) ;

  releaseBufPtr(env, buf, baseType, bufptr, offset);

  (*env)->SetIntField(env,stat, sourceID,status->MPI_SOURCE);
  (*env)->SetIntField(env,stat, tagID,status->MPI_TAG);
  (*env)->SetIntField(env,stat, indexID,index);

  free(reqs);

  return stat;
}

/*
 * Class:     mpi_Request
 * Method:    Testany
 * Signature: ([Lmpi/Request;Lmpi/Status;)Lmpi/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_Testany
  (JNIEnv *env, jclass jthis, jobjectArray array_of_request, jobject stat)
{
  int i,flag,index;
  int count=(*env)->GetArrayLength(env,array_of_request);
  MPI_Request *reqs=(MPI_Request*)calloc(count, sizeof(MPI_Request));
  MPI_Status *status =
      (MPI_Status *)((*env)->GetLongField(env,stat,stathandleID));

  for(i=0; i<count; i++)
    reqs[i]=(MPI_Request)((*env)->GetLongField(env,
                (*env)->GetObjectArrayElement(env,array_of_request,i),
                reqhandleID));

  MPI_Testany(count, reqs, &index,&flag, status);
  
  for(i=0; i<count; i++)
    (*env)->SetLongField(env,
                (*env)->GetObjectArrayElement(env,array_of_request,i),
                reqhandleID, (jlong) reqs[i]);

  free(reqs);

  if(flag) {
    jobject buf  = (*env)->GetObjectField(env, jthis, bufSaveID) ;
    void* bufptr = (void*) (*env)->GetLongField(env, jthis, bufptrSaveID) ;
    int baseType = (*env)->GetIntField(env, jthis, baseTypeSaveID) ;
    int offset   = (*env)->GetIntField(env, jthis, offsetSaveID) ;

    releaseBufPtr(env, buf, baseType, bufptr, offset);

    (*env)->SetIntField(env,stat, sourceID,status->MPI_SOURCE);
    (*env)->SetIntField(env,stat, tagID,status->MPI_TAG);
    (*env)->SetIntField(env,stat, indexID,index);

    return stat;
  }
  else
    return NULL;
}

/*
 * Class:     mpi_Request
 * Method:    waitall
 * Signature: ([Lmpi/Request;)[Lmpi/Status;
 */
JNIEXPORT jobjectArray JNICALL Java_mpi_Request_waitall
  (JNIEnv *env, jclass jthis, jobjectArray array_of_request)
{
  int i;
  int count=(*env)->GetArrayLength(env,array_of_request);
  MPI_Request *reqs=(MPI_Request*)calloc(2 * count, sizeof(MPI_Request));
  MPI_Request *reqs_ini = reqs + count ;
  MPI_Status *stas=(MPI_Status*)calloc(count, sizeof(MPI_Status));

  jclass status_class = (*env)->FindClass(env,"mpi/Status");
  jobjectArray array_of_status = 
    (*env)->NewObjectArray(env,count,status_class,NULL);

  /* Copy initial native requests in Java `array_of_request' to `reqs'. */

  for(i=0; i<count; i++) {
    reqs [i] = (MPI_Request)((*env)->GetLongField(env,
                  (*env)->GetObjectArrayElement(env,array_of_request,i),
                  reqhandleID));
    reqs_ini [i] = reqs [i] ;
  }

  MPI_Waitall(count, reqs, stas);

  for(i=0; i<count; i++) 
    if(reqs_ini [i] != MPI_REQUEST_NULL) {
      jobject req  = (*env)->GetObjectArrayElement(env,array_of_request,i) ;

      jobject buf  = (*env)->GetObjectField(env, req, bufSaveID) ;
      void* bufptr = (void*) (*env)->GetLongField(env, req, bufptrSaveID) ;
      int baseType = (*env)->GetIntField(env, req, baseTypeSaveID) ;
      int offset   = (*env)->GetIntField(env, req, offsetSaveID) ;

      jobject jstas = (*env)->NewObject(env,status_class,handleConstructorID);
      MPI_Status *status =
          (MPI_Status *)((*env)->GetLongField(env,jstas,stathandleID));

      /* Copy final native request to `array_of_request'.
         Release buffer elements... */

      (*env)->SetLongField(env, req, reqhandleID, (jlong) reqs[i]) ;

      releaseBufPtr(env, buf, baseType, bufptr, offset);

      /* Copy final native status to Java `array_of_status'... */

      *status = stas [i] ;

      (*env)->SetIntField(env,jstas,sourceID,status->MPI_SOURCE);
      (*env)->SetIntField(env,jstas,tagID,status->MPI_TAG);   

      (*env)->SetObjectArrayElement(env,array_of_status,i,jstas);

      /* Try not to create too many local references... */

      (*env)->DeleteLocalRef(env, req) ;
      (*env)->DeleteLocalRef(env, buf) ;
      (*env)->DeleteLocalRef(env, jstas) ;
    }
    
  free(reqs);
  free(stas);

  return array_of_status;
}


/*
 * Class:     mpi_Request
 * Method:    testall
 * Signature: ([Lmpi/Request;)[Lmpi/Status;
 */
JNIEXPORT jobjectArray JNICALL Java_mpi_Request_testall
  (JNIEnv *env, jclass jthis, jobjectArray array_of_request)
{
  int i,flag;
  int count=(*env)->GetArrayLength(env,array_of_request);
  MPI_Request *reqs=(MPI_Request*)calloc(2 * count, sizeof(MPI_Request));
  MPI_Request *reqs_ini = reqs + count ;
  MPI_Status *stas=(MPI_Status*)calloc(count, sizeof(MPI_Status));

  jclass status_class = (*env)->FindClass(env,"mpi/Status");
  jobjectArray array_of_status = 
    (*env)->NewObjectArray(env,count,status_class,NULL);

  /* Copy initial native requests in Java `array_of_request' to `reqs'. */

  for(i=0; i<count; i++) {
    reqs [i] =(MPI_Request)((*env)->GetLongField(env,
                 (*env)->GetObjectArrayElement(env,array_of_request,i),
                 reqhandleID)) ;
    reqs_ini [i] = reqs [i] ;
  }

  MPI_Testall(count, reqs, &flag, stas);

  if(flag)
    for(i=0; i<count; i++)
      if(reqs_ini [i] != MPI_REQUEST_NULL) {
        jobject req  = (*env)->GetObjectArrayElement(env,array_of_request,i) ;

        jobject buf  = (*env)->GetObjectField(env, req, bufSaveID) ;
        void* bufptr = (void*) (*env)->GetLongField(env, req, bufptrSaveID) ;
        int baseType = (*env)->GetIntField(env, req, baseTypeSaveID) ;
        int offset   = (*env)->GetIntField(env, req, offsetSaveID) ;

        jobject jstas = (*env)->NewObject(env,status_class,
                                          handleConstructorID);
        MPI_Status *status =
            (MPI_Status *)((*env)->GetLongField(env,jstas,stathandleID));

        /* Copy final native request to `array_of_request'.
           Release buffer elements... */

        (*env)->SetLongField(env, req, reqhandleID, (jlong) reqs[i]) ;

        releaseBufPtr(env, buf, baseType, bufptr, offset);

        /* Copy final native status to Java `array_of_status'... */

        *status = stas [i] ;

        (*env)->SetIntField(env,jstas,sourceID,status->MPI_SOURCE);
        (*env)->SetIntField(env,jstas,tagID,status->MPI_TAG);   

        (*env)->SetObjectArrayElement(env,array_of_status,i,jstas);

        /* Try not to create too many local references... */

        (*env)->DeleteLocalRef(env, req) ;
        (*env)->DeleteLocalRef(env, buf) ;
        (*env)->DeleteLocalRef(env, jstas) ;
      }
    
  free(reqs);
  free(stas);

  if(flag)
    return array_of_status ;
  else
    return NULL;
}

/*
 * Class:     mpi_Request
 * Method:    waitsome
 * Signature: ([Lmpi/Request;)[Lmpi/Status;
 */
JNIEXPORT jobjectArray JNICALL Java_mpi_Request_waitsome
  (JNIEnv *env, jclass jthis, jobjectArray array_of_request)
{
  int i;
  int incount=(*env)->GetArrayLength(env,array_of_request);
  MPI_Request *reqs=(MPI_Request*)calloc(incount, sizeof(MPI_Request));
  MPI_Status *stas=(MPI_Status*)calloc(incount, sizeof(MPI_Status));
  int *array_of_indices=(int*)calloc(incount,sizeof(int));
  int outcount;

  jclass status_class = (*env)->FindClass(env,"mpi/Status");
  jobjectArray array_of_status;

  /* Copy initial native requests in Java `array_of_request' to `reqs'. */

  for(i=0; i<incount; i++)
    reqs[i] = (MPI_Request)((*env)->GetLongField(env,
	         (*env)->GetObjectArrayElement(env,array_of_request,i),
                 reqhandleID));

  MPI_Waitsome(incount, reqs, &outcount, array_of_indices, stas); 

  if(outcount!=MPI_UNDEFINED) {
    array_of_status=(*env)->NewObjectArray(env,outcount,status_class,NULL);

    for(i=0; i<outcount; i++) {
      int index    = array_of_indices[i] ;

      jobject req  = (*env)->GetObjectArrayElement(env,array_of_request,
                                                   index) ;

      jobject buf  = (*env)->GetObjectField(env, req, bufSaveID) ;
      void* bufptr = (void*) (*env)->GetLongField(env, req, bufptrSaveID) ;
      int baseType = (*env)->GetIntField(env, req, baseTypeSaveID) ;
      int offset   = (*env)->GetIntField(env, req, offsetSaveID) ;

      jobject jstas = (*env)->NewObject(env,status_class,handleConstructorID);
      MPI_Status *status =
          (MPI_Status *)((*env)->GetLongField(env,jstas,stathandleID));

      /* Copy final native request to `array_of_request'.
         Release buffer elements... */

      (*env)->SetLongField(env, req, reqhandleID, (jlong) reqs[index]) ;

      releaseBufPtr(env, buf, baseType, bufptr, offset);

      /* Copy final native status to Java `array_of_status'... */

      *status = stas [i] ;

      (*env)->SetIntField(env,jstas,sourceID,status->MPI_SOURCE);
      (*env)->SetIntField(env,jstas,tagID,status->MPI_TAG);   
      (*env)->SetIntField(env,jstas,indexID, index);

      (*env)->SetObjectArrayElement(env,array_of_status,i,jstas);

      /* Try not to create too many local references... */

      (*env)->DeleteLocalRef(env, req) ;
      (*env)->DeleteLocalRef(env, buf) ;
      (*env)->DeleteLocalRef(env, jstas) ;
    }
  }

  free(reqs);
  free(stas);
  free(array_of_indices);

  if(outcount==MPI_UNDEFINED)
    return NULL;
  else
    return array_of_status;
}

/*
 * Class:     mpi_Request
 * Method:    testsome
 * Signature: ([Lmpi/Request;)[Lmpi/Status;
 */
JNIEXPORT jobjectArray JNICALL Java_mpi_Request_testsome
 (JNIEnv *env, jclass jthis, jobjectArray array_of_request)
{
  int i;
  int incount=(*env)->GetArrayLength(env,array_of_request);
  MPI_Request *reqs=(MPI_Request*)calloc(incount, sizeof(MPI_Request));
  MPI_Status *stas=(MPI_Status*)calloc(incount, sizeof(MPI_Status));
  int *array_of_indices=(int*)calloc(incount,sizeof(int));
  int outcount;

  jclass status_class = (*env)->FindClass(env,"mpi/Status");
  jobjectArray array_of_status;

  for(i=0; i<incount; i++) {
    reqs[i]=(MPI_Request)( (*env)->GetLongField(env,
        (*env)->GetObjectArrayElement(env,array_of_request,i),
        reqhandleID) );
  }

  MPI_Testsome(incount,reqs,&outcount,array_of_indices, stas);

  if(outcount!=MPI_UNDEFINED) {  
    array_of_status=(*env)->NewObjectArray(env,outcount,status_class,NULL);

    for(i=0; i<outcount; i++) {
      int index    = array_of_indices[i] ;

      jobject req  = (*env)->GetObjectArrayElement(env,array_of_request,
                                                   index) ;

      jobject buf  = (*env)->GetObjectField(env, req, bufSaveID) ;
      void* bufptr = (void*) (*env)->GetLongField(env, req, bufptrSaveID) ;
      int baseType = (*env)->GetIntField(env, req, baseTypeSaveID) ;
      int offset   = (*env)->GetIntField(env, req, offsetSaveID) ;

      jobject jstas = (*env)->NewObject(env,status_class,handleConstructorID);
      MPI_Status *status =
          (MPI_Status *)((*env)->GetLongField(env,jstas,stathandleID));

      /* Copy final native request to `array_of_request'.
         Release buffer elements... */

      (*env)->SetLongField(env, req, reqhandleID, (jlong) reqs[index]) ;

      releaseBufPtr(env, buf, baseType, bufptr, offset);

      /* Copy final native status to Java `array_of_status'... */

      *status = stas [i] ;

      (*env)->SetIntField(env,jstas,sourceID,status->MPI_SOURCE);
      (*env)->SetIntField(env,jstas,tagID,status->MPI_TAG);   
      (*env)->SetIntField(env,jstas,indexID, index);

      (*env)->SetObjectArrayElement(env,array_of_status,i,jstas);

      /* Try not to create too many local references... */

      (*env)->DeleteLocalRef(env, req) ;
      (*env)->DeleteLocalRef(env, buf) ;
      (*env)->DeleteLocalRef(env, jstas) ;
    }
  }

  free(reqs);
  free(stas);
  free(array_of_indices);
  
  if(outcount==MPI_UNDEFINED)
    return NULL;
  else
    return array_of_status;
}

/*
 * Things to do:
 *
 *   Should be able to cache a global reference to `status_class'.
 *       Doesn't work for some reason.  Why?
 *
 */

